;;;; -*- Mode: LISP; Syntax: ANSI-Common-Lisp; Base: 10 -*-

;;;; *************************************************************************
;;;; FILE IDENTIFICATION
;;;;
;;;; Name:          api.lisp
;;;; Purpose:       A Common Lisp front-end for working with CouchDb.
;;;; Programmer:    Nicolas Lamirault <nicolas.lamirault@gmail.com>
;;;;
;;;; This file, part of cl-couchdb, is Copyright (c) 2007 by Nicolas Lamirault
;;;;
;;;; cl-couchdb are granted the rights to distribute and use this software
;;;; as governed by the terms of the Lisp Lesser GNU Public License
;;;; (http://opensource.franz.com/preamble.html), also known as the LLGPL.
;;;;
;;;; *************************************************************************


(in-package :cl-couchdb)


(defclass couch ()
  ((host :initform nil
         :initarg :host
         :accessor couch-host
         :documentation "The hostname of the CouchDB server.")
   (port :initform nil
         :initarg :port
         :accessor couch-port
         :documentation "The port of the CouchDB server."))
  (:documentation "Wrapper to perform operations on a CouchDB server."))


(defun make-couch (host &optional (port *couchdb-port*))
  "Creates a new wrapper for a CouchDB server on HOST bind to PORT."
  (make-instance 'couch :host host :port port))


(defmethod print-object ((couch couch) stream)
  (if *print-couch*
      (with-slots (host port) couch
        (format stream "~&Couch: ~A ~A" host port))
      (print-unreadable-object (couch stream :type t :identity t))))


;; ------
;; Tools
;; ------


(defgeneric couch-perform (couch query &key parameters method content)
  (:documentation "Make a QUERY to the CouchDB server.
METHOD specify which HTTP method must be used.
PARAMETERS is an alist of name/value pairs.
CONTENT is a JSON data used when METHOD is a POST method.
Return two values : the HTTP code and the content of the anwser.
Throws a USOCKET:CONNECTION-REFUSED-ERROR if connection to the server
failed."))


(defmethod couch-perform ((couch couch) query &key parameters (method :get) content)
  (with-slots (host port) couch
    (let ((url (format nil "http://~A:~A/~A" host port query)))
      (when *debug*
        (format t "~&Couch : ~A: ~A~%" method url))
      (multiple-value-bind (body-stream status-code headers uri stream must-close)
          (if (equal :post method)
              (drakma:http-request url
                                   :method method
                                   :form-data t
                                   :external-format-out :UTF-8
                                   :content content)
              (drakma:http-request url
                                   :method method
                                   :content-type *content-type*
                                   :parameters parameters))
        (declare (ignore headers uri stream must-close))
        (values status-code body-stream)))))


(defmacro with-couchdb-server (couch &body body)
  "Macro which establish a connection to CouchDB server using COUCH
and executes BODY.
Throws a COUCH-CONNECTION-ERROR if connection to the server failed."
  `(handler-case
       ,@body
     (usocket:connection-refused-error (condition)
       (declare (ignore condition))
       (error 'couch-connection-error 
              :host (couch-host ,couch)
              :port (couch-port ,couch)))))


(defgeneric version (couch)
  (:documentation "Get information abouth the CouchDB server.
Throws a COUCH-REQUEST-ERROR if an error occurs during the HTTP request.
Throws a COUCH-CONNECTION-ERROR if connection to the server failed."))


(defmethod version ((couch couch))
  (with-couchdb-server couch
    (multiple-value-bind (code body)
        (couch-perform couch *version-uri*)
      (if (= code 200)
          body
          (error 'couch-request-error :code code :message body)))))


;; ------------
;; DB Protocol
;; ------------


(defgeneric list-db (couch)
  (:documentation "List the databases on the server.
Throws a COUCH-REQUEST-ERROR if an error occurs during the HTTP request."))


(defmethod list-db ((couch couch))
  (with-couchdb-server couch
    (multiple-value-bind (code body)
        (couch-perform couch *list-db-uri*)
      (if (= code 200)
          body
          (error 'couch-request-error :code code :message body)))))


(defgeneric create-db (couch db-name)
  (:documentation "Creates a new database DB-NAME on the server. A database
must be named with all lowercase characters (a-z), digits (0-9), or any
of these characters _$()+-/.
Throws a COUCH-DATABASE-ALREADY-EXISTS-ERROR is database already exists.
Throws a COUCH-REQUEST-ERROR if an error occurs during the HTTP request."))


(defmethod create-db ((couch couch) db-name)
  (with-couchdb-server couch
    (multiple-value-bind (code body)
        (couch-perform couch (format nil *create-db-uri* db-name) :method :put)
      (cond ((= code 201)
             body)
            ((= code 409)
             (error 'couch-database-already-exists-error
                    :code code :message db-name))
            (t
             (error 'couch-request-error :code code :message body))))))
    

(defgeneric delete-db (couch db-name)
  (:documentation "Deletes the database DB-NAME on the server.
Throws a COUCH-DATABASE-DOESNT-EXISTS-ERROR is database doesn't exists in
the database.
Throws a COUCH-REQUEST-ERROR if an error occurs during the HTTP request."))


(defmethod delete-db ((couch couch) db-name)
  (with-couchdb-server couch
    (multiple-value-bind (code body)
        (couch-perform couch (format nil *delete-db-uri* db-name) :method :delete)
      (cond ((= code 202)
             body)
            ((= code 404)
             (error 'couch-database-doesnt-exists-error
                    :code code :message db-name))
            (t
             (error 'couch-request-error :code code :message body))))))
    

(defgeneric info-db (couch db-name)
  (:documentation "Returns info about the database DB-NAME.
Throws a COUCH-DATABASE-DOESNT-EXISTS-ERROR is database doesn't exists in
the database.
Throws a COUCH-REQUEST-ERROR if an error occurs during the HTTP request."))


(defmethod info-db ((couch couch) db-name)
  (with-couchdb-server couch
    (multiple-value-bind (code body)
        (couch-perform couch (format nil *info-db-uri* db-name) :method :get)
      (cond ((= code 200)
             body)
            ((= code 404)
             (error 'couch-database-doesnt-exists-error
                    :code code :message db-name))
            (t
             (error 'couch-request-error :code code :message body))))))



;; ------------------
;; Document Protocol
;; ------------------


(defgeneric list-documents (couch db-name &optional reverse-p)
  (:documentation "Get a listing of all documents in a database DB-NAME.
Set REVERSE-P to T to reverse the order of the output.
Throws a COUCH-DATABASE-DOESNT-EXISTS-ERROR is database doesn't exists in
the database.
Throws a COUCH-REQUEST-ERROR if an error occurs during the HTTP request."))


(defmethod list-documents ((couch couch) db-name &optional reverse-p)
  (with-couchdb-server couch
    (multiple-value-bind (code body)
        (couch-perform couch
                       (if reverse-p
                           (format nil *list-docs-uri* db-name "?reverse=true")
                           (format nil *list-docs-uri* db-name "")))
      (cond ((= code 200)
             body)
            ((= code 404)
             (error 'couch-database-doesnt-exists-error
                    :code code :message db-name))
            (t
             (error 'couch-request-error :code code :message body))))))


(defgeneric open-document (couch db-name doc-id)
  (:documentation "Retrieve a document from a given database DB-NAME.
Throws a COUCH-DATABASE-DOESNT-EXISTS-ERROR is database doesn't exists in
the database.
Throws a COUCH-REQUEST-ERROR if an error occurs during the HTTP request."))


(defmethod open-document ((couch couch) db-name doc-id)
  (with-couchdb-server couch
    (multiple-value-bind (code body)
        (couch-perform couch (format nil *open-doc-uri* db-name doc-id))
      (cond ((= code 200)
             body)
            ((= code 404)
             (error 'couch-database-doesnt-exists-error
                    :code code :message db-name))
            (t
             (error 'couch-request-error :code code :message body))))))
 

(defgeneric save-document (couch db-name content &optional doc-id)
  (:documentation "Create or update a document to a given database DB-NAME.
If DOC-ID is set, update the existing document.
CONTENT is a an alist.
Throws a COUCH-DATABASE-DOESNT-EXISTS-ERROR is database doesn't exists in
the database.
Throws a COUCH-REQUEST-ERROR if an error occurs during the HTTP request."))


(defmethod save-document ((couch couch) db-name content &optional doc-id)
  (with-couchdb-server couch
    (multiple-value-bind (code body)
        (couch-perform couch (format nil *put-doc-uri* db-name) ;; doc-id)
                       :method :post
                       :content content)
      (cond ((= code 201)
             body)
            ((= code 404)
             (error 'couch-database-doesnt-exists-error
                    :code code :message db-name))
            (t
             (error 'couch-request-error :code code :message body))))))


(defgeneric delete-document (content db-name doc-id)
  (:documentation "Delete a document from a given database DB-NAME
with a server generated DOC-ID.
Throws a COUCH-DATABASE-DOESNT-EXISTS-ERROR is database doesn't exists in
the database.
Throws a COUCH-REQUEST-ERROR if an error occurs during the HTTP request."))


(defmethod delete-document ((couch couch) db-name doc-id)
  (with-couchdb-server couch
    (multiple-value-bind (code body)
        (couch-perform couch (format nil *delete-doc-uri* db-name doc-id)
                       :method :delete)
      (cond ((= code 202)
             body)
            ((= code 404)
             (error 'couch-database-doesnt-exists-error
                    :code code :message db-name))
            (t
             (error 'couch-request-error :code code :message body))))))


(defgeneric view-document (couch db-name doc-id &optional view-name)
  (:documentation ""))


(defmethod view-document ((couch couch) db-name doc-id &optional view-name)
  (with-couchdb-server couch
    (multiple-value-bind (code body)
        (couch-perform couch (format nil *view-doc-uri* db-name doc-id))
;;                      (if view-name
;;                          (format nil *view-doc-uri* db-name doc-id view-name)
;;                          (format nil *view-doc-uri* db-name doc-id "all_documents")))
      (cond ((= code 200)
             body)
            (t
             (error 'couch-request-error :code code :message body))))))


;; --------------
;; View Protocol
;; --------------


(defgeneric add-view (couch db-name view-name content)
  (:documentation "Add a new view named VIEW-NAME to the current database
DB-NAME.
CONTENT is the Javascript function to display document.
Throws a COUCH-DATABASE-DOESNT-EXISTS-ERROR is database doesn't exists in
the database.
Throws a COUCH-REQUEST-ERROR if an error occurs during the HTTP request."))


(defmethod add-view ((couch couch) db-name view-name content)
  (let ((name (concatenate 'string "_design_" view-name)))
    (save-document couch db-name content name)))


